"""
Apply ``coro`` to each emitted value allowing async processing

Usage:

>>> import asyncio
>>> from broqer import Value, Sink, op
>>> s = Value()

>>> async def delay_add(a):
...     print('Starting with argument', a)
...     await asyncio.sleep(0.015)
...     result = a + 1
...     print('Finished with argument', a)
...     return result

AsyncMode: CONCURRENT (is default)

>>> s.emit(0)
>>> _d = (s | op.MapAsync(delay_add)).subscribe(Sink())
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument 0
Starting with argument 1
Finished with argument 0
Finished with argument 1
>>> _d.dispose()

AsyncMode: INTERRUPT

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.INTERRUPT))
>>> _d = o.subscribe(Sink(print))
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.005))
Starting with argument 0
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument 1
Finished with argument 1
2
>>> _d.dispose()

AsyncMode: QUEUE

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.QUEUE))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
Starting with argument 1
Finished with argument 1
2
>>> _d.dispose()

AsyncMode: LAST

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.LAST))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> s.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
Starting with argument 2
Finished with argument 2
3
>>> _d.dispose()

AsyncMode: SKIP

>>> s.emit(0)
>>> o = (s | op.MapAsync(delay_add, mode=op.AsyncMode.SKIP))
>>> _d = o.subscribe(Sink(print))
>>> s.emit(1)
>>> s.emit(2)
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.04))
Starting with argument 0
Finished with argument 0
1
>>> _d.dispose()

Using error_callback:

>>> def cb(*e):
...     print('Got error')

>>> s.emit('abc')
>>> _d = (s | op.MapAsync(delay_add, error_callback=cb)).subscribe(Sink(print))
>>> asyncio.get_event_loop().run_until_complete(asyncio.sleep(0.02))
Starting with argument abc
Got error
>>> _d.dispose()
"""
import asyncio
from collections import deque
from enum import Enum
import sys
from functools import wraps
from typing import Any, MutableSequence, Optional  # noqa: F401

# pylint: disable=cyclic-import
from broqer.operator import Operator, OperatorFactory
from broqer import Publisher, default_error_handler, NONE


def build_coro(coro, unpack, *args, **kwargs):
    """ building a coroutine receiving one argument and call it curried
    with *args and **kwargs and unpack it (if unpack is set)
    """
    if unpack:
        async def _coro(value):
            return await coro(*args, *value, **kwargs)
    else:
        async def _coro(value):
            return await coro(*args, value, **kwargs)

    return _coro


class AsyncMode(Enum):
    """ AyncMode defines how to act when an emit happens while an scheduled
    coroutine is running """
    CONCURRENT = 1  # just run coroutines concurrent
    INTERRUPT = 2  # cancel running and call for new value
    QUEUE = 3  # queue the value(s) and call after coroutine is finished
    LAST = 4  # use last emitted value after coroutine is finished
    LAST_DISTINCT = 5  # like LAST but only when value has changed
    SKIP = 6  # skip values emitted during coroutine is running


class AppliedMapAsync(Operator):
    """ Apply ``coro`` to each emitted value allowing async processing

    :ivar scheduled: Publisher emitting the value when coroutine is actually
        started.
    """
    def __init__(self, publisher: Publisher,
                 coro_with_args, mode=AsyncMode.CONCURRENT,
                 error_callback=default_error_handler,
                 ) -> None:
        Operator.__init__(self, publisher)

        self._coro = coro_with_args
        self._mode = mode
        self._error_callback = error_callback

        # ._future is the reference to a running coroutine encapsulated as task
        self._future = None  # type: Optional[asyncio.Future]

        # ._last_emit is used for LAST_DISTINCT and keeps the last emit from
        # source publisher
        self._last_emit = NONE  # type: Any

        # .scheduled is a Publisher which is emitting the value of the source
        # publisher when the coroutine is actually started
        self.scheduled = Publisher()

        # queue is initialized with following sizes:
        # Mode:                size:
        # QUEUE                unlimited
        # LAST, LAST_DISTINCT  1
        # all others           no queue used
        self._queue = None  # type: Optional[MutableSequence]

        if mode in (AsyncMode.QUEUE, AsyncMode.LAST, AsyncMode.LAST_DISTINCT):
            maxlen = (None if mode is AsyncMode.QUEUE else 1)
            self._queue = deque(maxlen=maxlen)

    def emit(self, value: Any, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        # check if a coroutine is already running
        if self._future is not None:
            # append to queue if a queue is used in this mode
            if self._queue is not None:
                self._queue.append(value)
                return

            # cancel the future if INTERRUPT mode is used
            if self._mode is AsyncMode.INTERRUPT and \
                    not self._future.done():
                self._future.cancel()
            # in SKIP mode just do nothin with this emit
            elif self._mode is AsyncMode.SKIP:
                return

        # start the coroutine
        self._run_coro(value)

    def _future_done(self, future):
        """ Will be called when the coroutine is done """
        try:
            # notify the subscribers (except result is an exception or NONE)
            result = future.result()  # may raise exception
            if result is not NONE:
                Publisher.notify(self, result)  # may also raise exception
        except asyncio.CancelledError:
            return
        except Exception:  # pylint: disable=broad-except
            self._error_callback(*sys.exc_info())

        # check if queue is present and something is in the queue
        if self._queue:
            value = self._queue.popleft()

            # start the coroutine
            self._run_coro(value)
        else:
            self._future = None

    def _run_coro(self, value):
        """ Start the coroutine as task """

        # when LAST_DISTINCT is used only start coroutine when value changed
        if self._mode is AsyncMode.LAST_DISTINCT and \
                value == self._last_emit:
            self._future = None
            return

        # store the value to be emitted for LAST_DISTINCT
        self._last_emit = value

        # publish the start of the coroutine
        self.scheduled.notify(value)

        # build the coroutine
        coro = self._coro(value)

        # create a task out of it and add ._future_done as callback
        self._future = asyncio.ensure_future(coro)
        self._future.add_done_callback(self._future_done)


class MapAsync(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Apply ``coro(*args, value, **kwargs)`` to each emitted value allow
    async processing.

    :param coro: coroutine to be applied on emit
    :param \\*args: variable arguments to be used for calling coro
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked as (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for calling coro
    """
    def __init__(self, coro, *args, mode=AsyncMode.CONCURRENT,
                 error_callback=default_error_handler,
                 unpack: bool = False, **kwargs) -> None:
        self._coro = build_coro(coro, unpack, *args, **kwargs)
        self._mode = mode
        self._error_callback = error_callback
        self._unpack = unpack

    def apply(self, publisher: Publisher):
        return AppliedMapAsync(publisher,
                               coro_with_args=self._coro,
                               mode=self._mode,
                               error_callback=self._error_callback,
                               )


def build_map_async(coro=None, *,
                    mode: AsyncMode = AsyncMode.CONCURRENT,
                    error_callback=default_error_handler,
                    unpack: bool = False):
    """ Decorator to wrap a function to return a Map operator.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_map_async(coro):
        return MapAsync(coro, mode=mode, error_callback=error_callback,
                        unpack=unpack)

    if coro:
        return _build_map_async(coro)

    return _build_map_async


def build_map_async_factory(coro=None, *,
                            mode: AsyncMode = AsyncMode.CONCURRENT,
                            error_callback=default_error_handler,
                            unpack: bool = False):
    """ Decorator to wrap a coroutine to return a factory for MapAsync
        operators.

    :param coro: coroutine to be wrapped
    :param mode: behavior when a value is currently processed
    :param error_callback: error callback to be registered
    :param unpack: value from emits will be unpacked (*value)
    """
    _mode = mode

    def _build_map_async(coro):
        @wraps(coro)
        def _wrapper(*args, mode=None, **kwargs) -> MapAsync:
            if ('unpack' in kwargs) or ('error_callback' in kwargs):
                raise TypeError('"unpack" and "error_callback" has to '
                                'be defined by decorator')
            if mode is None:
                mode = _mode
            return MapAsync(coro, *args, mode=mode, unpack=unpack,
                            error_callback=error_callback, **kwargs)
        return _wrapper

    if coro:
        return _build_map_async(coro)

    return _build_map_async



import asyncio
import subprocess
import operator

from broqer import op

(op.FromPolling(1, subprocess.check_output, 'uptime')
 | op.Map(str, encoding='utf - 8')
 | op.Map(str.split, sep=', ')
 | op.Map(lambda v:v[0])
 | op.Sink(print)
 )

loop = asyncio.get_event_loop()
loop.run_until_complete(asyncio.sleep(10))



from collections import OrderedDict
from unittest.mock import Mock

import pytest

from broqer.op.bitwise import BitwiseCombineLatest, map_bit
from broqer import Publisher, NONE, Sink
from tests import helper_multi, helper_single

test_vector = [
    # init, bit_value_map, input_vector, output_vector
    (0, [(0, True), (1, False), (4, True)],
        [(NONE, True, NONE)],
        [17, 19]),
    (0, [(0, True), (1, NONE), (4, True)],
        [(NONE, True, NONE)],
        [17, 19]),
    (~0, [(0, True)],
        [(True,), (False,), (True,)],
        [~0, ~0, ~1, ~0]),
    (0xAA, [(0, False), (1, False), (7, True)],
        [(NONE, True, NONE), (True, True, True), (False, NONE, NONE), (False, False, False)],
        (0xA8, 0xAA, 0xAB, 0xAA, 0x28))
]
@pytest.mark.parametrize('method', [helper_multi.check_get_method, helper_multi.check_subscription, helper_multi.check_dependencies])
@pytest.mark.parametrize('init,bit_value_map,input_vector,output_vector', test_vector)
def test_bitwise_combine_latest(method, init, bit_value_map, input_vector, output_vector):
    publisher_bit_mapping = OrderedDict([(Publisher(v), b) for b, v in bit_value_map])
    i_vector = [tuple(v for k, v in bit_value_map)] + input_vector

    operator = BitwiseCombineLatest(publisher_bit_mapping, init)

    method(operator, i_vector, output_vector)


def test_bitwise_uninitialized_publishers():
    m = Mock()
    p = Publisher()
    b = BitwiseCombineLatest({p: 0})
    b.subscribe(Sink(m))
    m.assert_called_once_with(0)

test_vector = [
    # bit_index, input_vector, output_vector
    (0, [0, 1, 2, 3, ~0], [False, True, False, True, True]),
    (128, [~0, 0, 1<<128, (1<<128)-1], [True, False, True, False])
]
@pytest.mark.parametrize('method', [helper_single.check_get_method, helper_single.check_subscription, helper_single.check_dependencies])
@pytest.mark.parametrize('bit_index, input_vector,output_vector', test_vector)
def test_map_bit(method, bit_index, input_vector, output_vector):
    operator = map_bit(bit_index)

    method(operator, input_vector, output_vector)



import asyncio
import pytest
from unittest import mock

from broqer import NONE, Sink, Publisher, op
from broqer.op import Throttle

from .eventloop import VirtualTimeEventLoop


@pytest.yield_fixture()
def event_loop():
    loop = VirtualTimeEventLoop()
    yield loop
    loop.close()


@pytest.mark.asyncio
async def test_throttle_errorhandler(event_loop):
    from broqer import default_error_handler

    p = Publisher()
    mock_sink = mock.Mock()
    mock_error_handler = mock.Mock()

    default_error_handler.set(mock_error_handler)

    throttle = p | op.Throttle(0.1)
    disposable = throttle.subscribe(Sink(mock_sink))

    mock_sink.side_effect = (None, ZeroDivisionError('FAIL'))

    # test error_handler
    p.notify(1)
    await asyncio.sleep(0.05, loop=event_loop)
    mock_sink.assert_called_once_with(1)
    p.notify(2)
    await asyncio.sleep(0.1, loop=event_loop)
    mock_error_handler.assert_called_once_with(ZeroDivisionError, mock.ANY, mock.ANY)
    mock_sink.assert_has_calls((mock.call(1), mock.call(2)))

    mock_sink.reset_mock()


@pytest.mark.asyncio
async def test_throttle_unsubscribe(event_loop):
    p = Publisher()
    mock_sink = mock.Mock()

    throttle = p | op.Throttle(0.1)
    disposable = throttle.subscribe(Sink(mock_sink))

    # test subscription and unsubscribe
    p.notify(2)
    mock_sink.assert_called_once_with(2)

    await asyncio.sleep(0.05, loop=event_loop)
    mock_sink.reset_mock()

    disposable.dispose()
    await asyncio.sleep(0.1, loop=event_loop)

    # dispose must not emit anything
    mock_sink.assert_not_called()

    p.notify(3)

    await asyncio.sleep(0.1, loop=event_loop)

    # after dispose was called, p.notify must not emit to mock_sink
    mock_sink.assert_not_called()


@pytest.mark.asyncio
async def test_throttle_reset(event_loop):
    p = Publisher()
    mock_sink = mock.Mock()

    throttle = p | op.Throttle(0.1)
    disposable = throttle.subscribe(Sink(mock_sink))

    p.notify(1)
    await asyncio.sleep(0.05, loop=event_loop)
    throttle.reset()
    p.notify(3)

    await asyncio.sleep(0.05, loop=event_loop)

    # reset is called after "1" was emitted
    mock_sink.assert_has_calls((mock.call(1), mock.call(3)))

    ## wait until initial state is set and reset mock
    await asyncio.sleep(0.1, loop=event_loop)
    mock_sink.reset_mock()

    p.notify(1)
    await asyncio.sleep(0.05, loop=event_loop)
    p.notify(2)
    throttle.reset()
    p.notify(3)

    await asyncio.sleep(0.05, loop=event_loop)

    # reset is called after "1" was emitted, and while "2" was hold back,
    #   therefore "1" and "3" are emitted, but "2" is ignored
    mock_sink.assert_has_calls((mock.call(1), mock.call(3)))

    disposable.dispose()


@pytest.mark.parametrize('emit_sequence, expected_emits', [
    (((0, 0), (0.05, 1), (0.4, 2), (0.6, 3), (0.2, 4), (0.2, 5)),
     (mock.call(0), mock.call(2), mock.call(3), mock.call(5))),
    (((0.001, 0), (0.6, 1), (0.5, 2), (0.05, 3), (0.44, 4)),
     (mock.call(0), mock.call(1), mock.call(2), mock.call(4))),
])
@pytest.mark.asyncio
async def test_throttle(event_loop, emit_sequence, expected_emits):
    p = Publisher()
    mock_sink = mock.Mock()

    throttle = p | op.Throttle(0.5)
    disposable = throttle.subscribe(Sink(mock_sink))

    mock_sink.assert_not_called()

    for item in emit_sequence:
        await asyncio.sleep(item[0], loop=event_loop)
        p.notify(item[1])

    await asyncio.sleep(0.5, loop=event_loop)

    mock_sink.assert_has_calls(expected_emits)


def test_argument_check():
    with pytest.raises(ValueError):
        Throttle(-1)



"""
Apply ``function(*args, value, **kwargs)`` to each emitted value

Usage:

>>> from broqer import Value, op, Sink
>>> s = Value()

>>> mapped_publisher = s | op.Map(lambda v:v*2)
>>> _disposable = mapped_publisher.subscribe(Sink(print))

>>> s.emit(1)
2
>>> s.emit(-1)
-2
>>> s.emit(0)
0
>>> _disposable.dispose()

Also possible with additional args and kwargs:

>>> import operator
>>> mapped_publisher = s | op.Map(operator.add, 3)
>>> _disposable = mapped_publisher.subscribe(Sink(print))
3
>>> s.emit(100)
103
>>> _disposable.dispose()

>>> _disposable = (s | op.Map(print, 'Output:')).subscribe(\
                                                Sink(print, 'EMITTED'))
Output: 100
EMITTED None
>>> s.emit(1)
Output: 1
EMITTED None
"""
from functools import partial, wraps
from typing import Any, Callable

from broqer import Publisher, NONE
from broqer.publisher import TValue
from broqer.operator import Operator, OperatorFactory


class AppliedMap(Operator):
    """ Map object applied to publisher (see Map) """
    def __init__(self, publisher: Publisher, function: Callable[[Any], Any],
                 unpack: bool = False) -> None:
        """ Special care for return values:
              - return `None` (or nothing) if you don't want to return a result
              - return `None, ` if you want to return `None`
              - return `(a, b), ` to return a tuple as value
              - every other return value will be unpacked
        """

        Operator.__init__(self, publisher)
        self._function = function
        self._unpack = unpack

    def get(self) -> TValue:
        if self._subscriptions:
            return self._state

        value = self._orginator.get()  # type: TValue

        if value is NONE:
            return value

        if self._unpack:
            assert isinstance(value, (list, tuple))
            return self._function(*value)

        return self._function(value)

    def emit(self, value: TValue, who: Publisher) -> None:
        if who is not self._orginator:
            raise ValueError('Emit from non assigned publisher')

        if self._unpack:
            assert isinstance(value, (list, tuple))
            result = self._function(*value)
        else:
            result = self._function(value)

        if result is not NONE:
            return Publisher.notify(self, result)

        return None


class Map(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ Apply ``function(*args, value, **kwargs)`` to each emitted value.

    :param function: function to be applied for each emit
    :param \\*args: variable arguments to be used for calling function
    :param unpack: value from emits will be unpacked (\\*value)
    :param \\*\\*kwargs: keyword arguments to be used for calling function
    """
    def __init__(self, function: Callable[[Any], Any],
                 *args, unpack: bool = False, **kwargs) -> None:
        self._function = partial(function, *args, **kwargs)
        self._unpack = unpack

    def apply(self, publisher: Publisher):
        return AppliedMap(publisher, self._function, self._unpack)


def build_map(function: Callable[..., None] = None, *,
              unpack: bool = False):
    """ Decorator to wrap a function to return a Map operator.

    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_map(function):
        return Map(function, unpack=unpack)

    if function:
        return _build_map(function)

    return _build_map


def build_map_factory(function: Callable[[Any], Any] = None,
                      unpack: bool = False):
    """ Decorator to wrap a function to return a factory for Map operators.

    :param function: function to be wrapped
    :param unpack: value from emits will be unpacked (*value)
    """
    def _build_map(function: Callable[[Any], Any]):
        @wraps(function)
        def _wrapper(*args, **kwargs) -> Map:
            if 'unpack' in kwargs:
                raise TypeError('"unpack" has to be defined by decorator')
            return Map(function, *args, unpack=unpack, **kwargs)
        return _wrapper

    if function:
        return _build_map(function)

    return _build_map



import asyncio
import pytest

from broqer import Publisher, OnEmitFuture

from .eventloop import VirtualTimeEventLoop

@pytest.yield_fixture()
def event_loop():
    loop = VirtualTimeEventLoop()
    yield loop
    loop.close()

def test_publisher():
    p = Publisher(1)
    future = OnEmitFuture(p, timeout=1, loop=asyncio.get_event_loop())

    assert future.result() == 1

    p.notify(2)
    assert future.result() == 1

@pytest.mark.asyncio
async def test_timeout(event_loop):
    p = Publisher()
    future = OnEmitFuture(p, timeout=0.01)
    await asyncio.sleep(0.05, loop=event_loop)

    with pytest.raises(asyncio.TimeoutError):
        future.result()

@pytest.mark.asyncio
async def test_cancel(event_loop):
    p = Publisher()
    future = OnEmitFuture(p, timeout=0.01)
    future.cancel()
    p.notify(1)

    with pytest.raises(asyncio.CancelledError):
        future.result()

def test_wrong_source():
    p = Publisher()
    on_emit_future = OnEmitFuture(p)

    with pytest.raises(ValueError):
        on_emit_future.emit(0, who=Publisher())


""" Implementing PollPublisher """
import asyncio
from typing import Callable, Type, Any, Optional

from broqer.publisher import Publisher, TValue, SubscriptionDisposable
from broqer.subscriber import Subscriber


class PollPublisher(Publisher):
    """ A PollPublisher is periodically calling a given function and is using
    the returned value to notify the subscribers. When no subscriber is
    present, the polling function is not called.

    :param poll_cb: Function to be called
    :param interval: Time in seconds between polling calls
    """
    def __init__(self, poll_cb: Callable[[], Any], interval: float, *,
                 type_: Type[TValue] = None):
        Publisher.__init__(self, type_=type_)
        self.poll_cb = poll_cb
        self.interval = interval
        self._poll_handler = None  # type: Optional[asyncio.TimerHandle]

    def subscribe(self, subscriber: Subscriber,
                  prepend: bool = False) -> SubscriptionDisposable:
        if not self._subscriptions:
            # call poll_cb once to set internal state and schedule a _poll call
            self._state = self.poll_cb()
            loop = asyncio.get_event_loop()
            assert self._poll_handler is None, '_poll_handler already assigned'
            self._poll_handler = loop.call_later(self.interval, self._poll)

        return Publisher.subscribe(self, subscriber, prepend)

    def unsubscribe(self, subscriber: 'Subscriber') -> None:
        Publisher.unsubscribe(self, subscriber)
        if not self._subscriptions and self._poll_handler:
            # when no subscription is left, cancel next _poll and reset state
            self._poll_handler.cancel()
            self._poll_handler = None
            self.reset_state()

    def _poll(self):
        assert self._poll_handler is not None

        value = self.poll_cb()
        Publisher.notify(self, value)

        loop = asyncio.get_event_loop()
        self._poll_handler = loop.call_later(self.interval, self._poll)

    def notify(self, value: TValue) -> None:
        """ PollPublisher does not support .notify calls """
        raise ValueError(self.notify.__doc__)



""" Concat enables concating operators """
from broqer.operator import OperatorFactory, Publisher


class Concat(OperatorFactory):  # pylint: disable=too-few-public-methods
    """ This class is generator a new operator by concatenation of other
    operators.

    :param operators: the operators to concatenate
    """
    def __init__(self, *operators):
        self._operators = operators

    def apply(self, publisher: Publisher) -> Publisher:
        # concat each operator in the following step
        orginator = publisher

        for operator in self._operators:
            orginator = operator.apply(orginator)

        # the source publisher is the last operator in the chain
        return orginator



from unittest import mock

import pytest

from broqer import Publisher, NONE, Sink


def check_get_method(operator, input_vector, output_vector):
    input_value, output_value = input_vector[0], output_vector[0]

    publishers = operator.dependencies

    for p, v in zip(publishers, input_value):
            p.get = mock.MagicMock(return_value=v)

    with pytest.raises(ValueError):
        operator.emit(input_value[0], who=None)

    assert operator.get() == output_value

    disposable = operator.subscribe(Sink())

    for p, v in zip(publishers, input_value):
        if v is not NONE:
            operator.emit(v, who=p)  # simulate emit on subscribe

    for p, v in zip(publishers, input_value):
        p.get = None  # .get should not be called when operator has a subscription

    assert operator.get() == output_value  # this should retrieve the internal state

    # after unsubscription it should remove the internal state and retrieve it
    # directly from orignator publisher via get
    disposable.dispose()

    for p, v in zip(publishers, input_value):
        p.get = mock.MagicMock(return_value=v)

    assert operator.get() == output_value

    for p in publishers:
        p.get.assert_called_once_with()


def check_subscription(operator, input_vector, output_vector):
    m = mock.Mock()

    publishers = operator.dependencies

    # subscribe operator to publisher
    disposable = operator.subscribe(Sink(m))

    for p in publishers:
        assert p.subscriptions == (operator,)  # now it should be subscribed

    assert operator.dependencies == publishers

    # test emit on subscription
    if output_vector[0] is not NONE:
        m.assert_called_once_with(output_vector[0])

    m.reset_mock()

    # test input_vector
    for input_value, output_value in zip(input_vector[1:], output_vector[1:]):
        for p, v in zip(publishers, input_value):
            if v is not NONE:
                p.notify(v)

        if output_value is NONE:
            m.assert_not_called()
        else:
            m.assert_called_with(output_value)
        m.reset_mock()

    # test input_vector with unsubscriptions between
    disposable.dispose()

    assert operator.dependencies == publishers

    for p, v in zip(publishers, input_vector[0]):
        p.notify(v)

    disposable = operator.subscribe(Sink(m))

    for input_value, output_value in zip(input_vector, output_vector):
        m.reset_mock()

        for p, v in zip(publishers, input_value):
            if v is not NONE:
                p.notify(v)

        if output_value is NONE:
            m.assert_not_called()
        else:
            m.assert_called_with(output_value)


def check_dependencies(operator, *_):
    publishers = operator.dependencies

    for p in publishers:
        assert p.subscriptions == ()  # operator should not be subscribed yet

    assert operator.subscriptions == ()

    # subscribe to operator
    disposable1 = operator.subscribe(Sink())

    assert p.subscriptions == (operator,)  # operator should now be subscriped
    assert operator.subscriptions == (disposable1.subscriber,)

    # second subscribe to operator
    disposable2 = operator.subscribe(Sink(), prepend=True)

    assert p.subscriptions == (operator,)
    assert operator.subscriptions == (disposable2.subscriber, disposable1.subscriber)

    # remove first subscriber
    disposable1.dispose()

    assert p.subscriptions == (operator,)
    assert operator.subscriptions == (disposable2.subscriber,)

    # remove second subscriber
    disposable2.dispose()

    assert p.subscriptions == ()  # operator should now be subscriped
    assert operator.subscriptions == ()




""" Enables accessing bits directly. Provides BitwiseCombineLatest to gather
publishers to bit indicies and evaluate a value from them.

map_bit builds a Publisher which is mapping to a specific bit on another
Publisher.
"""
from typing import Any, Dict  # noqa: F401

# pylint: disable=cyclic-import
from broqer import Publisher, Subscriber, NONE, SubscriptionDisposable
from broqer.op import build_map_factory

from broqer.operator import MultiOperator


class BitwiseCombineLatest(MultiOperator):
    """ Bitwise combine the latest emit of multiple publishers and emit the
    combination. If a publisher is not emitting or is not defined for a bit,
    the init value will be used.

    :param bit_publisher_mapping: dictionary with bit index as key and source
                                  publisher as value
    :param init: optional init value used for undefined bits (or initial state)
    """
    def __init__(self, publisher_bit_mapping: Dict, init: int = 0) -> None:
        MultiOperator.__init__(self, *publisher_bit_mapping)

        self._init = init
        self._missing = set(self._orginators)
        self._publisher_bit_mapping = publisher_bit_mapping

    def subscribe(self, subscriber: 'Subscriber',
                  prepend: bool = False) -> SubscriptionDisposable:
        disposable = MultiOperator.subscribe(self, subscriber, prepend)

        if self._missing:
            self._missing.clear()
            if self._state is NONE:
                Publisher.notify(self, self._init)
            else:
                Publisher.notify(self, self._state)

        return disposable

    def unsubscribe(self, subscriber: Subscriber) -> None:
        MultiOperator.unsubscribe(self, subscriber)
        if not self._subscriptions:
            self._missing.update(self._orginators)
            self._state = NONE

    def get(self):
        if self._subscriptions:
            return self._state

        state = self._init

        for publisher, bit_index in self._publisher_bit_mapping.items():
            value = publisher.get()

            if value is NONE:
                continue

            if value:
                state |= 1 << bit_index
            else:
                state &= ~(1 << bit_index)

        return state

    def emit(self, value: Any, who: Publisher) -> None:
        if all(who is not p for p in self._orginators):
            raise ValueError('Emit from non assigned publisher')

        # remove source publisher from ._missing
        self._missing.discard(who)

        # evaluate
        bit_index = self._publisher_bit_mapping[who]

        if self._state is NONE:
            self._state = self._init

        if value:
            self._state |= 1 << bit_index
        else:
            self._state &= ~(1 << bit_index)

        if self._missing:
            return None

        return Publisher.notify(self, self._state)


@build_map_factory()
def map_bit(bit_index, value):
    """ Provide value of a specific bit """
    return bool(value & (1 << bit_index))
